/*
Decoder.m - SMS Framework
Copyright (C) 2005, Tim Ellis

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#import "SMS.h"
#import "Decoder.h"

@implementation Decoder

- (SMSObject*)decode:(NSString*)raw {
	NSString *servicetoaoct, *sendertoaoct;
	SMSObject *sms = [[SMSObject alloc] init];
	int smscinfolen, senderaddlen, offset = 0;
	
	smscinfolen = [self decodeLength:raw offset:0];
	servicetoaoct = [NSString stringWithString:[self decodeHexOctet2Bin:raw offset:2]];
	[sms setSmscInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithInt:[self decodeToaType:servicetoaoct]], @"type",
		[NSNumber numberWithInt:[self decodeToaNpi:servicetoaoct]], @"npi", 
		[NSString stringWithString:[self decodeService:raw smscInfoLength:smscinfolen]], @"number", nil]];

	// 00 is for a SMS-DELIVER PDU so ignore anything else
	[sms setFirstOctetInfo:[NSDictionary dictionaryWithDictionary:[self decodeFirstOctet:[self decodeHexOctet2Bin:raw offset:((smscinfolen * 2) + 2)]]]];
	if([(NSString*)[[sms firstOctetInfo] objectForKey:@"mti"] isEqual:@"00"] == FALSE) return sms;
	
	senderaddlen = [self decodeLength:raw offset:((smscinfolen * 2) + 4)];
	sendertoaoct = [NSString stringWithString:[self decodeHexOctet2Bin:raw offset:((smscinfolen * 2) + 6)]];
	if(senderaddlen % 2 == 1) senderaddlen++;
	[sms setSenderInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithInt:[self decodeToaType:sendertoaoct]], @"type", 
		[NSNumber numberWithInt:[self decodeToaNpi:sendertoaoct]], @"npi", 
		[NSString stringWithString:[self decodeSender:raw senderAdLen:senderaddlen smscInfoLength:smscinfolen]], @"number", nil]];
	
	[sms setTpDcsInfo:[self decodeTpDcs:raw offset:(10 + senderaddlen + (smscinfolen * 2))]];
	[sms setTimestamp:[self decodeTimestamp:raw senderAdLen:senderaddlen smscInfoLength:smscinfolen]];
	
	if([[sms firstOctetInfo] objectForKey:@"udhi"] == [NSNumber numberWithInt:1])
		offset = [self decodeLength:raw offset:(28 + senderaddlen + (smscinfolen * 2))];

	[sms setMessageAndHeader:[NSDictionary dictionaryWithDictionary:[self decodeMessage:raw senderAdLen:senderaddlen 
		smscInfoLength:smscinfolen offset:offset alphabet:[[[sms tpDcsInfo] objectForKey:@"msgalphabet"] intValue]]]];
	if([[sms firstOctetInfo] objectForKey:@"udhi"] == [NSNumber numberWithInt:1]) {
		[sms setMultipartInfo:[self decodeHeaderInformationMultipart:[[sms messageAndHeader] objectForKey:@"header"]]];
	}
	return sms;
}

- (NSDictionary*)decodeFirstOctet:(NSString*)oct {
	NSNumber * rp, * udhi, * sri, * mms;
	if([oct characterAtIndex:0] == '1') rp = [NSNumber numberWithInt:1];
	else rp = [NSNumber numberWithInt:0];
	if([oct characterAtIndex:1] == '1') udhi = [NSNumber numberWithInt:1];
	else udhi = [NSNumber numberWithInt:0];
	if([oct characterAtIndex:2] == '1') sri = [NSNumber numberWithInt:1];
	else sri = [NSNumber numberWithInt:0];
	if([oct characterAtIndex:5] == '1') mms = [NSNumber numberWithInt:1];
	else mms = [NSNumber numberWithInt:0];
	return [NSDictionary dictionaryWithObjectsAndKeys:rp, @"replypath", udhi, @"udhi", sri, @"sri", mms, @"mms", 
		[oct substringWithRange:NSMakeRange(6, 2)], @"mti", nil];
}

- (NSDictionary*)decodeMessage:(NSString*)raw senderAdLen:(int)al smscInfoLength:(int)se offset:(int)offset alphabet:(int)alpha {
	NSString *header;
	NSMutableString *message;
	int octhex, cl, a;
	NSString *rawsms = [NSString stringWithString:[raw substringWithRange:NSMakeRange((28 + al + (se * 2)), ([raw length] - (28 + al + (se * 2))))]];
	if(alpha == kTpDcsAlphaDefault) {
		// Message type is default 7 bit
		octhex = [rawsms length] / 2;
		NSString * binsms = [NSString stringWithString:[self decodeMessageHex2bin:octhex hexString:rawsms]];
		NSString * binsms2 = [NSString stringWithString:[self decodeMessageDec8bbin7bbin:octhex eightBin:binsms]];
		if(offset != 0) {
			offset = (offset * 8) + 8;
			header = [NSString stringWithString:[binsms substringWithRange:NSMakeRange(0, offset)]];
			if(offset % 7 != 0) {
				do {
					offset++;
				} while(offset % 7 != 0);
			}
			message = [NSMutableString stringWithString:[self decodeMessage7BBin2Message:
				[binsms2 substringWithRange:NSMakeRange(offset, [binsms2 length] - offset)]]];
		} else {
			message = [NSMutableString stringWithString:[self decodeMessage7BBin2Message:binsms2]];
			header = [NSString stringWithString:@""];
		}
	} else if(alpha == kTpDcsAlphaUcs2) {
		// Message type is ucs2
		message = [NSMutableString string];
		if(offset != 0) offset = (offset + 1) * 2;
		for(cl = 0; cl < ([rawsms length] - offset) / 4; cl++) {
			a = strtol([[rawsms substringWithRange:NSMakeRange((cl * 4) + offset, 4)] UTF8String], NULL, 16);
			[message appendFormat:@"%C", a];
		}
		header = [self decodeMessageHex2bin:[[rawsms substringWithRange:NSMakeRange(0, offset)] length] / 2 hexString:[rawsms substringWithRange:NSMakeRange(0, offset)]];
	} else if(alpha == kTpDcsAlphaData) {
		// Message type is 8 bit data
		return nil;	
	} else if(alpha == kTpDcsAlphaReserved) {
		// Message type is reserved
		return nil;	
	}
	return [NSDictionary dictionaryWithObjectsAndKeys:header, @"header", message, @"data", nil];
}

- (NSString*)decodeService:(NSString*)raw smscInfoLength:(int)se {
	NSMutableString *returnString = [NSMutableString string];
	int al, start = 5;
	for(al = 0; al < se - 1; al++) {
		[returnString appendFormat:@"%C", [raw characterAtIndex:(start + (al * 2))]];
		if(al == se - 2) {
			if([raw characterAtIndex:(start + (al * 2) - 1)] != 'F') [returnString appendFormat:@"%C", [raw characterAtIndex:(start + (al * 2) - 1)]];
		} else [returnString appendFormat:@"%C", [raw characterAtIndex:(start + (al * 2) - 1)]];
	}	
	return returnString;
}

- (NSString*)decodeSender:(NSString*)raw senderAdLen:(int)al smscInfoLength:(int)se {
	int cl = (se * 2) + 9, b, odd = al % 2;
	NSMutableString *temp = [NSMutableString string];
	unichar t;
 	for(b = 0; b < ((al + odd) / 2); b++) {
		[temp appendFormat:@"%C", [raw characterAtIndex:cl]];
		if(odd == 1 && b != ((al + odd) / 2) - 1) [temp appendFormat:@"%C", [raw characterAtIndex:(cl - 1)]];
		else if(odd == 0) [temp appendFormat:@"%C", [raw characterAtIndex:(cl - 1)]];
		cl = cl + 2;
	}
	cl = 0;
	for(b = 0; b < al; b++) {
		t = [temp characterAtIndex:b];
		if ((t == 'A') || (t == 'B') || (t == 'C') || (t == 'D') || (t == 'E') || (t == 'F')) cl++;
		if(b == al - 1 && t == 'F') {
			cl--;
			[temp deleteCharactersInRange:NSMakeRange(b, 1)];
		}
	}
	if(cl == 0) return temp;
	else {
		[temp setString:@""];
		cl = 22;
		for(b = 0; b < al; b++) {
			[temp appendFormat:@"%C", [raw characterAtIndex:cl]];
			cl++;
		}
		return [self decodeMessage7BBin2Message:[self decodeMessageDec8bbin7bbin:([temp length] / 2) 
													eightBin:[self decodeMessageHex2bin:([temp length] / 2) hexString:temp]]];
	}
}

- (NSString*)decodeMessageDec8bbin7bbin:(int)octhex eightBin:(NSString*)binsms {
	int first = 1, bit = 1, co = 0, b;
	NSString *firstbin = [NSString string], *lastbin = [NSString string], *interbin = [NSString string];
	NSMutableString *finalbin = [NSMutableString string], *tempbin = [NSMutableString string];
	do {
		[tempbin setString:@""];
		for(b = 0; b < 8; b++) {
			[tempbin appendFormat:@"%C", [binsms characterAtIndex:((co * 8) + b)]];
			if (b == 7) co++;
		}
		if (bit == 1) {
			if(first == 0) [finalbin appendString:lastbin];
			else {
				lastbin = [NSString string];
				first = 0;
			}
		}
		interbin = [NSString stringWithString:[tempbin substringWithRange:NSMakeRange(0, bit)]];
		firstbin = [NSString stringWithString:[tempbin substringWithRange:NSMakeRange(bit, (8 - bit))]];
		[finalbin appendString:firstbin];
		if(bit != 1) [finalbin appendString:lastbin];
		lastbin = [NSString stringWithString:interbin];
		if (bit == 7) bit = 1;
		else bit++;
	} while (co < octhex);
	if(octhex % 8 != 0) [finalbin appendString:lastbin];
	return finalbin;
}

- (NSString*)decodeMessage7BBin2Message:(NSString*)binMessage {
	NSMutableString *messbin = [NSMutableString string], *message = [NSMutableString string];
	int theLength, count = 0, b, chara, rem;
	theLength = [binMessage length];
	rem = theLength % 7;	
	do {
		[messbin setString:@""];
		for(b = 0; b <=6; b++) {
			if(b == 0) [messbin appendString:@"0"];
			[messbin appendFormat:@"%C", [binMessage characterAtIndex:count]];
			count++;
		}
		chara = strtol([messbin UTF8String], NULL, 2);
		[message appendFormat:@"%c", chara];
	} while (count < (theLength - rem));
	return message;
}

- (NSDate*)decodeTimestamp:(NSString*)raw senderAdLen:(int)al smscInfoLength:(int)se {
	int cl = 13 + al + (se * 2), b, ans = 0;
	NSMutableString *dateString = [NSMutableString stringWithString:[[[NSDate date] description] substringWithRange:NSMakeRange(0, 2)]];
 	for(b = 0; b < 7; b++) {
		if(b != 6) [dateString appendFormat:@"%C%C", [raw characterAtIndex:cl], [raw characterAtIndex:cl - 1]];
		cl = cl + 2;
		if(b < 2) [dateString appendString:@"-"];
		if((b == 2) || (b == 5)) [dateString appendString:@" "];
		if((b == 3) || (b == 4)) [dateString appendString:@":"];
		if(b == 5) {
			if([[self decodeMessageHex2bin:1 hexString:[NSString stringWithFormat:@"%C%C", [raw characterAtIndex:cl - 2], [raw characterAtIndex:cl - 3]]] 
				characterAtIndex:7] == '1') [dateString appendString:@"-"];
			else [dateString appendString:@"+"];
		}
		if(b == 6) {
			ans = strtol([[NSString stringWithFormat:@"%C%C", [raw characterAtIndex:cl - 2], [raw characterAtIndex:cl - 3]] cString], NULL, 10);
		}
	}
	int seconds = ans * 900;
	int temp = ((seconds / 60) / 60) - (((seconds / 60) / 60) % 1);
	if(temp < 10) [dateString appendFormat:@"%@%d", @"0", temp];
	else [dateString appendFormat:@"%d", temp];
	seconds = seconds - ((temp * 60) * 60);
	temp = (seconds / 60) - ((seconds / 60) % 1);
	if(temp < 10) [dateString appendFormat:@"%@%d", @"0", temp];
	else [dateString appendFormat:@"%d", temp];
	return [NSDate dateWithString:dateString];
}

- (int)decodeToaNpi:(NSString*)octet {
	NSString * subString = [octet substringWithRange:NSMakeRange(4, 4)];
	if([subString isEqualToString:@"0001"]) return kToaNpiISDNTelephone;
	if([subString isEqualToString:@"0011"]) return kToaNpiData;
	if([subString isEqualToString:@"0100"]) return kToaNpiTelex;
	if([subString isEqualToString:@"1000"]) return kToaNpiNational;
	if([subString isEqualToString:@"1001"]) return kToaNpiPrivate;
	if([subString isEqualToString:@"1010"]) return kToaNpiErmes;
	if([subString isEqualToString:@"1111"]) return kToaNpiReserved;
	else return kToaNpiUnknown;
}

- (int)decodeToaType:(NSString*)octet {
	NSString * subString = [octet substringWithRange:NSMakeRange(1, 3)];
	if([subString isEqualToString:@"001"]) return kToaTypeInternational;
	if([subString isEqualToString:@"010"]) return kToaTypeNational;
	if([subString isEqualToString:@"011"]) return kToaTypeNetwork;
	if([subString isEqualToString:@"100"]) return kToaTypeSubscriber;
	if([subString isEqualToString:@"101"]) return kToaTypeAlphanumeric;
	if([subString isEqualToString:@"110"]) return kToaTypeAbbbreviated;
	if([subString isEqualToString:@"111"]) return kToaTypeReserved;
	else return kToaTypeUnknown;
}

- (NSString*)decodeMessageHex2bin:(int)numOct hexString:(NSString*)hexStr {
	NSMutableString *returnString = [NSMutableString string];
	int cl, co, rem, ans;
	char tempbin[9], tm[3];
	for(cl = 0; cl < numOct; cl++) {
		for(co = 0; co < 9; co++) tempbin[co] = '\0';
		snprintf(tm, 3, [[hexStr substringWithRange:NSMakeRange((cl * 2), 2)] cString]);
		ans = strtol(tm, NULL, 16);
		for(co = 7; co >= 0; co--) {
			rem = ans % 2;
			if(rem == 0) tempbin[co] = '0';
			else if ((rem == 1) || (rem == -1)) {
				tempbin[co] = '1';
				ans--;
			}
			ans = ans / 2;
		}
		[returnString appendString:[NSString stringWithCString:tempbin]];
	}
	return returnString;
}

- (int)decodePdufirstCheckTpmti:(NSString*)firstoctet {
	if(([firstoctet characterAtIndex:6] == '0') && ([firstoctet characterAtIndex:7] == '0')) return 0;
	else return -1;
}

- (int)decodeLength:(NSString*)raw offset:(int)theOffset {return strtol([[raw substringWithRange:NSMakeRange(theOffset, 2)] cString], NULL, 16);}

- (NSString*)decodeHexOctet2Bin:(NSString*)raw offset:(int)theOffset {
	return [self decodeMessageHex2bin:1 hexString:[raw substringWithRange:NSMakeRange(theOffset, 2)]];
}

- (NSDictionary*)decodeHeaderInformationMultipart:(NSString*)theHeader {
	int a, i, headerLength = [theHeader length] - 8;
	NSString *header = [NSString stringWithString:[theHeader substringWithRange:NSMakeRange(8, headerLength)]];
	NSNumber *smsid, *total, *partno;
	for(i = 0; i < headerLength / 8; i++) {
		if([[header substringWithRange:NSMakeRange(i * 8, 8)] isEqualToString:@"00000000"]) {
			a = strtol([[header substringWithRange:NSMakeRange(i * 8 + 16, 8)] UTF8String], NULL, 2);
			smsid = [[NSNumber alloc] initWithInt:a];
			a = strtol([[header substringWithRange:NSMakeRange(i * 8 + 24, 8)] UTF8String], NULL, 2);
			total = [[NSNumber alloc] initWithInt:a];
			a = strtol([[header substringWithRange:NSMakeRange(i * 8 + 32, 8)] UTF8String], NULL, 2);
			partno = [[NSNumber alloc] initWithInt:a];
			return [NSDictionary dictionaryWithObjectsAndKeys:smsid, @"id", total, @"total", partno, @"partno", nil];
		} else if([[header substringWithRange:NSMakeRange(i * 8, 8)] isEqualToString:@"00001111"]) {
			a = strtol([[header substringWithRange:NSMakeRange(i * 8 + 16, 16)] UTF8String], NULL, 2);
			smsid = [[NSNumber alloc] initWithInt:a];
			a = strtol([[header substringWithRange:NSMakeRange(i * 8 + 32, 8)] UTF8String], NULL, 2);
			total = [[NSNumber alloc] initWithInt:a];
			a = strtol([[header substringWithRange:NSMakeRange(i * 8 + 40, 8)] UTF8String], NULL, 2);
			partno = [[NSNumber alloc] initWithInt:a];
			return [NSDictionary dictionaryWithObjectsAndKeys:smsid, @"id", total, @"total", partno, @"partno", nil];
		} else {
			a = strtol([[header substringWithRange:NSMakeRange(i * 8 + 8, 8)] UTF8String], NULL, 2);
			i = i + a;
		}
	}
	return nil;
}

- (NSDictionary*)decodeTpDcs:(NSString*)raw offset:(int)theOffset {
	int dcsType, alphabet = kInvalid, messageClass = kInvalid, isCompressed = kInvalid, classValid = kInvalid, indActivity = kInvalid, indType = kInvalid;
	NSString *octet = [NSString stringWithString:[self decodeMessageHex2bin:1 hexString:[raw substringWithRange:NSMakeRange(theOffset, 2)]]];
	NSString *alpha;
	if([[octet substringWithRange:NSMakeRange(0, 2)] isEqualToString:@"00"]) {
		// General Data Coding
		dcsType = kTpDcsGdc;
		messageClass = [self decodeTpDcsMessageClass:octet];
		alpha = [NSString stringWithString:[octet substringWithRange:NSMakeRange(4, 2)]];
		if([alpha isEqualToString:@"00"]) alphabet = kTpDcsAlphaDefault;
		else if([alpha isEqualToString:@"10"]) alphabet = kTpDcsAlphaUcs2;
		else if([alpha isEqualToString:@"01"]) alphabet = kTpDcsAlphaData;
		else alphabet = kTpDcsAlphaReserved;
		alpha = [NSString stringWithString:[octet substringWithRange:NSMakeRange(2, 1)]];
		if([alpha isEqualToString:@"0"]) isCompressed = kTpDcsMsgUncompressed;
		else if([alpha isEqualToString:@"1"]) isCompressed = kTpDcsMsgCompressed;
		alpha = [NSString stringWithString:[octet substringWithRange:NSMakeRange(3, 1)]];
		if([alpha isEqualToString:@"0"]) classValid = kTpDcsGdcClassInvalid;
		else if([alpha isEqualToString:@"1"]) classValid = kTpDcsGdcClassValid;
	} else if([[octet substringWithRange:NSMakeRange(0, 4)] isEqualToString:@"1111"]) {
		// Data coding/message class
		dcsType = kTpDcsMwigData;
		messageClass = [self decodeTpDcsMessageClass:octet];
		alpha = [NSString stringWithString:[octet substringWithRange:NSMakeRange(5, 1)]];
		if([alpha isEqualToString:@"0"]) alphabet = kTpDcsAlphaDefault;
		else if([alpha isEqualToString:@"1"]) alphabet = kTpDcsAlphaData;
	} else if([[octet substringWithRange:NSMakeRange(0, 4)] isEqualToString:@"1101"]) {
		// Message Waiting Indication Group: Store Message Default Alphabet
		dcsType = kTpDcsMwigStoreDefault;
		alphabet = kTpDcsAlphaDefault;
		indActivity = [self decodeTpDcsIndicationActivity:octet];
		indType = [self decodeTpDcsMessageClass:octet];
		isCompressed = kTpDcsMsgUncompressed;
	} else if([[octet substringWithRange:NSMakeRange(0, 4)] isEqualToString:@"1110"]) {
		// Message Waiting Indication Group: Store Message UCS2 Alphabet
		dcsType = kTpDcsMwigStoreUcs2;
		alphabet = kTpDcsAlphaUcs2;
		indActivity = [self decodeTpDcsIndicationActivity:octet];
		indType = [self decodeTpDcsMessageClass:octet];
		isCompressed = kTpDcsMsgUncompressed;
	} else if([[octet substringWithRange:NSMakeRange(0, 4)] isEqualToString:@"1100"]) {
		// Message Waiting Indication Group: Discard Message
		dcsType = kTpDcsMwigDiscard;
		indActivity = [self decodeTpDcsIndicationActivity:octet];
		indType = [self decodeTpDcsMessageClass:octet];
	} else {
		// Reserved coding groups
		dcsType = kTpDcsReserved;
	}
	
	return [NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithInt:dcsType], @"dcstype", [NSNumber numberWithInt:alphabet], @"msgalphabet",
		[NSNumber numberWithInt:messageClass], @"msgclass", [NSNumber numberWithInt:isCompressed], @"msgcompression", 
		[NSNumber numberWithInt:classValid], @"msgclassvalidity", [NSNumber numberWithInt:indActivity], @"indactivity",
		[NSNumber numberWithInt:indType], @"indtype", nil];
}

- (int)decodeTpDcsMessageClass:(NSString*)octet {
	NSString * bits = [NSString stringWithString:[octet substringWithRange:NSMakeRange(6, 2)]];
	if([bits isEqualToString:@"00"]) return kTpDcsMsgClassAlert;
	else if([bits isEqualToString:@"01"]) return kTpDcsMsgClassMe;
	else if([bits isEqualToString:@"10"]) return kTpDcsMsgClassSim;
	else return kTpDcsMsgClassTe;
}

- (int)decodeTpDcsIndicationActivity:(NSString*)octet {
	if([[octet substringWithRange:NSMakeRange(4, 1)] isEqualToString:@"0"]) return kTpDcsMwigIndInactive;
	else return kTpDcsMwigIndActive;
}

- (int)decodeTpDcsIndicationType:(NSString*)octet {
	NSString * bits = [NSString stringWithString:[octet substringWithRange:NSMakeRange(6, 2)]];
	if([bits isEqualToString:@"00"]) return kTpDcsMwigIndTypeVmail;
	else if([bits isEqualToString:@"01"]) return kTpDcsMwigIndTypeFax;
	else if([bits isEqualToString:@"10"]) return kTpDcsMwigIndTypeEmail;
	else return kTpDcsMwigIndTypeOther;
}

@end
